#!/usr/bin/env python
import os
import shutil
import difflib
import pprint
import web_minify.processor
from web_minify.files import generate_file_list
from web_minify.handlers import HandlerManager

import logging

logger = logging.getLogger(__name__)

base_path: str = os.path.abspath(os.path.dirname(__file__))

input_path: str = os.path.abspath(os.path.join(base_path, "input/"))
output_base_path: str = os.path.abspath(os.path.join(base_path, "output/"))
expected_base_path: str = os.path.abspath(os.path.join(base_path, "expected/"))


try:
    from colorama import Fore, Back, Style, init

    init()
except ImportError:  # fallback so that the imported classes always exist

    class ColorFallback:
        __getattr__ = lambda self, name: ""

    Fore = Back = Style = ColorFallback()


def color_diff(diff):
    for line in diff:
        if line.startswith("+"):
            yield Fore.GREEN + line + Fore.RESET
        elif line.startswith("-"):
            yield Fore.RED + line + Fore.RESET
        elif line.startswith("^"):
            yield Fore.BLUE + line + Fore.RESET
        else:
            yield line


def _test_handlers():
    settings = {"mode": "test", "input": "", "output": "", "verbose": False}
    hm = HandlerManager(settings)
    assert hm is not None
    extensions = hm.get_supported_extensions()
    assert len(extensions) > 0
    for ext in extensions:
        handler = hm.get_handler_by_extension(ext)
        assert handler is not None
        handler.is_binary()
        handler.extensions()
        handler.process("")
    inexistant_h = hm.get_handler_by_extension("inexistant")
    assert inexistant_h is None


def test_minify_regressions():
    return regressions_work("minify")


def _test_beautify_regressions():
    return regressions_work("beautify")


def regressions_work(mode):
    output_path = os.path.join(output_base_path, mode)
    expected_path = os.path.join(expected_base_path, mode)
    expected_files = generate_file_list(expected_path)

    def clear_output():
        shutil.rmtree(output_path, ignore_errors=True)

    # 1. Clear old result if any
    # clear_output()
    # 2. Run tool on input
    # fmt:off
    settings = {
        "mode": mode
        , "input": input_path
        , "output": output_path
        , "verbose": False
        , "overwrite": True
        , "force": True
        #, "disable_type_css": True
        #, "disable_type_html": True
        #, "disable_type_sass": True
        #, "disable_type_png": True
        #, "disable_type_jpeg": True
        #, "disable_type_svg": True
    }
    # fmt:on
    processor = web_minify.processor.Processor(settings)
    res, msg = processor.sanity_checks()
    if not res:
        logger.warning(msg)
    assert res
    processor.process_files()
    # 3. Compare output with expected
    # output_files = generate_file_list(output_path)
    # logger.warning(pprint.pformat(expected_files))
    # logger.warning(pprint.pformat(output_files))
    # logger.warning(f"expected_path: {expected_path}")
    # logger.warning(f"output_path: {output_path}")
    errors = 0
    for expected_file in expected_files:
        # logger.warning(f"EXPECTED FILE: {expected_file}")
        rel = expected_file[expected_file.startswith(expected_path) and len(expected_path) + 1 :]
        # logger.warning(f"REL: {rel}")
        output_file = os.path.join(output_path, rel)
        # logger.warning(f"OUTPUTY FILE: {output_file}")
        assert os.path.exists(output_file)
        output_is_binary = output_file.endswith((".png", ".jpeg", ".jpg"))
        expected_is_binary = expected_file.endswith((".png", ".jpeg", ".jpg"))
        output_content = ""
        with open(output_file, "rb") if output_is_binary else open(output_file, "r", encoding="utf-8") as output_handle:
            output_content = output_handle.read()
        expected_content = ""
        with open(expected_file, "rb") if expected_is_binary else open(expected_file, "r", encoding="utf-8") as expected_handle:
            expected_content = expected_handle.read()
        if output_content != expected_content:
            logger.warning(f"--- Output did not match expected for {rel}:")
            diff = difflib.ndiff(output_content.splitlines(1), expected_content.splitlines(1))
            diff = color_diff(diff)
            logger.error("\n".join(diff))
            errors += 1
    assert errors == 0

    # 4. Clear result
    # clear_output()



import sys
import logging
import json

logger = logging.getLogger(__name__)


class Handler:
    """json handler class"""

    def __init__(self, settings: dict):
        self.settings = settings

    @classmethod
    def is_binary(self):
        return False

    @classmethod
    def extensions(self):
        return ["json"]

    @classmethod
    def name(self):
        return "json"

    def _process(self, raw: str, do_format: bool):
        try:
            parsed = json.loads(raw)
            format_indent = 4
            return json.dumps(parsed, indent=format_indent if do_format else None, sort_keys=True), None
        except json.JSONDecodeError as jde:
            return f"Error decoding json: {jde}", None
        except Exception as e:
            return f"Unknown error processing json: {e}", None

    def process(self, raw: str, name: str = None):
        mode = self.settings.get("mode")
        if "beautify" == mode:
            return self._process(raw, do_format=True)
        elif "minify" == mode:
            return self._process(raw, do_format=False)
        else:
            logger.warning(f"Unsupported mode '{mode}' for {self.name()} handler, skipping...")
            return raw, None



from .scour import scourString

import logging

logger = logging.getLogger(__name__)


def svg_minify(svg, settings: dict):
    """Minify SVG main function."""
    svg = scourString(in_string=svg, options=settings)
    return svg, None


class Handler:
    """SVG handler class"""

    def __init__(self, settings: dict):
        self.settings = settings

    @classmethod
    def is_binary(self):
        return False

    @classmethod
    def extensions(self):
        return ["svg"]

    @classmethod
    def name(self):
        return "svg"

    def process(self, raw: str, name: str = None):
        mode = self.settings.get("mode")
        if "beautify" == mode:
            # TODO: Implement
            return raw, None
        elif "minify" == mode:
            return svg_minify(raw, self.settings)
        else:
            logger.warning(f"Unsupported mode '{mode}' for {self.name()} handler, skipping...")
            return raw, None



import os
import pkg_resources
from .files import read_file

import logging

logger = logging.getLogger(__name__)

__version__ = None
version_filename = "VERSION"

wm_path: str = os.path.abspath(os.path.join(os.path.dirname(__file__), "../"))
wm_version_file = f"{wm_path}/{version_filename}"

places_looked = []

if pkg_resources.resource_exists(__name__, version_filename):
    __version__ = pkg_resources.resource_string(__name__, version_filename).decode("utf-8").strip()

if not __version__ and os.path.exists(wm_version_file):
    __version__ = read_file(wm_version_file)

if not __version__:
    logger.warning(f"No version found in wm_path={wm_path}, wm_version_file={wm_version_file}")
    __version__ = "0.0.0"



#!/usr/bin/env python3
"""CSS Minifier functions for CSS-HTML-JS-Minify."""

from .variables import EXTENDED_NAMED_COLORS, CSS_PROPS_TEXT
from .css import css_minify

__all__ = ("css_minify", "condense_semicolons")

import logging

logger = logging.getLogger(__name__)


class Handler:
    """CSS handler class"""

    def __init__(self, settings: dict):
        self.settings = settings

    @classmethod
    def is_binary(self):
        return False

    @classmethod
    def extensions(self):
        return ["css"]

    @classmethod
    def name(self):
        return "css"

    def process(self, raw: str, name: str = None):
        mode = self.settings.get("mode")
        if "beautify" == mode:
            # TODO: Implement
            return raw, None
        elif "minify" == mode:
            return css_minify(raw, self.settings)
        else:
            logger.warning(f"Unsupported mode '{mode}' for {self.name()} handler, skipping...")
            return raw, None



from io import StringIO

from calmjs.parse import io
from calmjs.parse import rules
from calmjs.parse.exceptions import ECMASyntaxError
from calmjs.parse.lexers.es5 import Lexer
from calmjs.parse.parsers.es5 import parse
from calmjs.parse.unparsers.es5 import Unparser

import logging

logger = logging.getLogger(__name__)


class Crimper:
    def __init__(self, settings):
        self.settings = settings

    def crimp(self, raw, mangle=True, obfuscate=True, indent_width=4, drop_semi=True, encoding="utf-8"):
        enabled_rules = [rules.minify(drop_semi=drop_semi or mangle)]
        if obfuscate or mangle:
            enabled_rules.append(rules.obfuscate(reserved_keywords=Lexer.keywords_dict.keys()))
        # if pretty:
        #    enabled_rules.append(rules.indent(indent_str=' ' * indent_width))
        printer = Unparser(rules=enabled_rules)
        out = raw
        with StringIO(raw) as input_stream:
            with StringIO() as output_stream:
                try:
                    io.write(printer, io.read(parse, input_stream), output_stream)
                except ECMASyntaxError as e:
                    return raw, [f"ECMASyntaxError: {e}"]
                except (IOError, OSError) as e:
                    return raw, [f"IOError: {e}"]
                except UnicodeDecodeError as e:
                    return raw, [f"UnicodeDecodeError (read) error: {e}"]
                except UnicodeEncodeError as e:
                    return raw, [f"UnicodeEncodeError (write) error: {e}"]
                except Exception as e:
                    logger.exception(
                        f"Unknown error: {e}",
                    )
                    return raw, [f"Unknown error: {e}"]
                # no need to close any streams as they are callables and that the io
                # read/write functions take care of that.
                out = output_stream.getvalue()
        return out, None



import pprint

from jsbeautifier.javascript.options import BeautifierOptions
from jsbeautifier.javascript.beautifier import Beautifier

import esprima

from .jscodegen import CodeGenerator
from .obfuscator import obfuscate

import logging

logger = logging.getLogger(__name__)


generator = CodeGenerator(indent=2)

beautifier = Beautifier()
beautifier_opts = BeautifierOptions()


def js2ast(js: str, name: str = ""):
    ret = {}
    # Trivial reject, parser does not handle empty input
    if len(js.strip()) <= 0:
        logger.warn("Empty input, skipping parsing")
        return ret
    options = {"tolerant": True, "comment": True}
    # parser = esprima.esprima.Parser(js, options=options, delegate=None)

    # ret = parser.parseScript()
    try:
        ret = esprima.parseScript(js, options=options).toDict()
    except esprima.error_handler.Error as e:
        logger.exception("ESPRIMA PARSING FAILED: {e}")
    if "errors" in ret:
        for e in ret.get("errors"):
            logger.error(f"ESPRIMA PARSING ERROR: {e}")
    # logger.info("JS2AST --------------------------")
    # logger.info(pprint.pformat(ret))
    return ret


def ast2js(ast: dict, name: str = ""):
    if not ast:
        logger.warn("Empty input, skipping generation")
        return ""
    ret = generator.generate(ast, name)
    # logger.info("AST2JS --------------------------")
    # logger.info(pprint.pformat(ret))
    return ret


def beautify(js: str, name: str = ""):
    # ret = beautifier.beautify(js, beautifier_opts)
    # ret = dummy_js
    ret = js
    # logger.info(f"SOURCE ({name})--------------------------")
    # logger.info(pprint.pformat(ret))
    return ret


class Handler:
    """JS handler class"""

    def __init__(self, settings: dict):
        self.settings = settings

    @classmethod
    def is_binary(self):
        return False

    @classmethod
    def extensions(self):
        return ["js"]

    @classmethod
    def name(self):
        return "js"

    def process(self, raw: str, name: str = None):
        mode = self.settings.get("mode")
        if "beautify" == mode:
            js = beautify(raw, name)
            # logger.info(f"returning beautified {js}")
            return js, None
        elif "minify" == mode:
            # js = ast2js(js2ast(beautify(raw, name),name),name)
            js = raw
            return js, None
        else:
            logger.warning(f"Unsupported mode '{mode}' for {self.name()} handler, skipping...")
            return raw, None



# https://engineeringblog.yelp.com/2017/06/making-photos-smaller.html

import logging

logger = logging.getLogger(__name__)


def jpeg_minify(jpeg, settings: dict):
    """Minify JPEG main function."""
    # TODO: Implement this
    # print(f"JPEG WAS CALLED FOR {len(jpeg)}!!!!! #############################################")
    return jpeg, None


class Handler:
    """JPEG handler class"""

    def __init__(self, settings: dict):
        self.settings = settings

    @classmethod
    def is_binary(self):
        return True

    @classmethod
    def extensions(self):
        return ["jpeg", "jpg"]

    @classmethod
    def name(self):
        return "jpeg"

    def process(self, raw: str, name: str = None):
        mode = self.settings.get("mode")
        if "beautify" == mode:
            # TODO: Implement
            return raw, None
        elif "minify" == mode:
            return jpeg_minify(raw, self.settings)
        else:
            logger.warning(f"Unsupported mode '{mode}' for {self.name()} handler, skipping...")
            return raw, None



import importlib
import pkgutil

import logging


logger = logging.getLogger(__name__)


class HandlerManager:
    def __init__(self):
        self.handler_extensions = []
        self.handler_names = []
        self.handler_classes = []
        self.handlers_by_name = {}
        self.handlers_by_extension = {}
        self._enumerate_handlers()

    def _enumerate_handlers(self):
        self.handlers = {}
        for importer, modname, is_package in pkgutil.iter_modules(__path__):
            imported_mod = importlib.import_module(f"web_minify.handlers.{modname}")
            if not imported_mod:
                logger.warning(f"Could not import module '{modname}', skipping...")
                continue
            handler_class = getattr(imported_mod, "Handler")
            if not handler_class:
                logger.warning(f"Could not find Handler class for '{modname}', skipping...")
                continue
            if not isinstance(handler_class, type):
                logger.warning(f"Handler symbol is not class for '{modname}', skipping...")
                continue
            handler_extensions = handler_class.extensions()
            if not handler_extensions:
                logger.warning(f"Handler does not support any extensions for '{modname}', skipping...")
                continue
            handler_name = handler_class.name()
            self.handler_classes.append(handler_class)
            self.handler_names.append(handler_name)
            self.handler_extensions.extend(handler_extensions)

    def prepare_handlers(self, settings):
        self.handlers = {}
        for handler_class in self.handler_classes:
            handler = handler_class(settings)
            if not handler:
                logger.warning(f"Could not instanciate handler for '{handler_class.name()}', skipping...")
                continue
            self.handlers_by_name[handler.name()] = handler
            for extension in handler.extensions():
                # logger.warning(f"Configuring handler {name} for .{extension}")
                self.handlers_by_extension[extension] = handler

    def get_handler_by_extension(self, extension):
        return self.handlers_by_extension.get(extension, None)

    def get_handler_by_name(self, name):
        return self.handlers_by_name.get(name, None)

    def get_supported_extensions(self):
        return self.handler_extensions

    def get_handler_names(self):
        return self.handler_names



import sys
import logging
from jinja2 import Environment

logger = logging.getLogger(__name__)

jinja_env = Environment()


def jinja_minify(jinja, settings: dict):
    """Minify jinja main function. Does not minify, rather detects errors (lint)"""
    try:
        jinja_env.parse(jinja)
    except Exception as e:
        logger.error("Jinja validator: {e}")
    return jinja.strip(), None


class Handler:
    """jinja2 handler class"""

    def __init__(self, settings: dict):
        self.settings = settings

    @classmethod
    def is_binary(self):
        return False

    @classmethod
    def extensions(self):
        return ["jinja", "jinja2", "j2"]

    @classmethod
    def name(self):
        return "jinja"

    def process(self, raw: str, name: str = None):
        mode = self.settings.get("mode")
        if "beautify" == mode:
            # TODO: Implement
            return raw, None
        elif "minify" == mode:
            return jinja_minify(raw, self.settings)
        else:
            logger.warning(f"Unsupported mode '{mode}' for {self.name()} handler, skipping...")
            return raw, None



from .css import css_minify

import logging

logger = logging.getLogger(__name__)

have_sass = False

try:
    import sass

    have_sass = True
except Exception:
    logger.warning("Could not import sass", exc_info=True)
    logger.warning("NOTE: To install sass, please remember to use  use pip install libsass instead of pip install sass")


class Handler:
    """SASS handler class"""

    def __init__(self, settings: dict):
        self.settings = settings

    @classmethod
    def is_binary(self):
        return False

    @classmethod
    def extensions(self):
        return ["sass"]

    @classmethod
    def name(self):
        return "sass"

    def process(self, raw: str, name: str = None):
        mode = self.settings.get("mode")
        if "beautify" == mode:
            # TODO: Implement
            return raw, None
        elif "minify" == mode:
            return sass_minify(raw, self.settings)
        else:
            logger.warning(f"Unsupported mode '{mode}' for {self.name()} handler, skipping...")
            return raw, None


def sass_minify(scss, settings: dict):
    """Minify SASS main function."""
    if have_sass:
        try:
            css = sass.compile(string=scss)
        except sass.CompileError as ce:
            return (None, [f"SASS {ce}"])
        except Exception as e:
            return (None, [f"Error: Could not process SASS {e}"])
    css, css_errors = css_minify(css=css, settings=settings)
    return css, css_errors



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""HTML Minifier functions for CSS-HTML-JS-Minify."""

import re
import logging
from bs4 import BeautifulSoup

from .html import html_minify

logger = logging.getLogger(__name__)


def html_beautify(raw, settings):
    return BeautifulSoup(raw, "html.parser").prettify(), None


class Handler:
    """HTML handler class"""

    def __init__(self, settings: dict):
        self.settings = settings

    @classmethod
    def is_binary(self):
        return False

    @classmethod
    def extensions(self):
        return ["html", "htm"]

    @classmethod
    def name(self):
        return "html"

    def process(self, raw: str, name: str = None):
        mode = self.settings.get("mode")
        if "beautify" == mode:
            return html_beautify(raw, self.settings)
        elif "minify" == mode:
            return html_minify(raw, self.settings)
        else:
            logger.warning(f"Unsupported mode '{mode}' for {self.name()} handler, skipping...")
            return raw, None



import os
import pprint
import datetime
import time
from .files import path_is_in_path, generate_file_list, determine_file_extension
from .handlers import HandlerManager
from .time import human_delta
from .modes import Mode

# from multiprocessing import Pool, cpu_count
from pathos.multiprocessing import ProcessingPool as Pool

import logging

logger = logging.getLogger(__name__)

now = datetime.datetime.now()

gzip_suffix = ".gz"


class Processor:
    def __init__(self, settings: dict):
        self.settings = settings
        self.input = settings.get("input", None)
        self.output = settings.get("output", None)
        self.input_is_dir = self.input and os.path.isdir(self.input)
        self.input_exists = self.input and os.path.exists(self.input)
        self.output_is_dir = self.output and os.path.isdir(self.output)
        self.output_exists = self.output and os.path.exists(self.output)

        self.mode: Mode = Mode(settings.get("mode", "minify"))
        self.format: bool = settings.get("format", False)
        self.overwrite: bool = settings.get("overwrite", False)
        self.on_change: bool = settings.get("on_change", False)
        self.prefix: bool = settings.get("prefix", False)
        self.gzip: bool = settings.get("gzip", False)
        self.hash: bool = settings.get("hash", False)
        self.verbose: bool = settings.get("verbose", False)
        self.force: bool = settings.get("force", False)
        self.dry_run: bool = settings.get("dry_run", False)
        self.nproc: int = settings.get("nproc", 1)

        self.handlers = HandlerManager()
        self.handlers.prepare_handlers(self.settings)

        self.valid_extensions = self.handlers.get_supported_extensions()
        self.handler_names = self.handlers.get_handler_names()

        if self.verbose:
            logger.info("### Settings:")
            logger.info(pprint.pformat(self.settings))
            logger.info("### Handlers:")
            logger.info(pprint.pformat(self.handler_names))
            logger.info("### Valid extensions:")
            logger.info(pprint.pformat(self.valid_extensions))
        self.pool = None

    def sanity_checks(self):
        valid_modes = list(Mode)
        if not self.mode in valid_modes:
            return (False, f"The specified mode {self.mode} was not valid. NOTE: Valid modes are {pprint.pformat(valid_modes)}")
        if not self.input or not self.input_exists:
            return (False, f"The input specified '{self.input}' did not exist. Input must be an existing directory or file")
        if not self.overwrite and self.on_change:
            return (False, f"On-change will not have an effect so long as overwrite is not enabled")
        if not self.output and not self.overwrite and not self.format:
            return (False, f"Only input '{self.input}' was specified. Without a setting for 'output', 'overwrite' and/or 'format' all processing will fail")
        if path_is_in_path(self.input, self.output):
            return (False, f"The output '{self.output}' is a subpath of input '{self.input}'")
        if path_is_in_path(self.output, self.input):
            return (False, f"The input '{self.input}' is a subpath of output '{self.output}'")
        return True, None

    def process_file(self, input_path: str, output_path: str):
        if self.verbose:
            logger.info(f"Processing '{input_path}' to '{output_path}'")
        if not input_path:
            return False, False, False, ["No input path specified"]
        if not output_path:
            return False, False, False, ["No output path specified"]
        input_mtime = os.path.getmtime(input_path)
        output_mtime = None
        if os.path.isfile(output_path):
            if not self.overwrite:
                return False, False, False, ["Trying to overwrite without overwrite enabled"]
            if self.on_change:
                output_mtime = os.path.getmtime(output_path)
                if not self.overwrite and not self.force and (input_mtime == output_mtime):
                    return True, False, True, ["File not changed, skipping"]
                elif not self.overwrite and not self.force and (output_mtime > input_mtime):
                    return True, False, True, [f"Destination file newer than input, skipping ({output_mtime} >= {input_mtime})"]
        else:
            output_dir = os.path.dirname(output_path)
            if not os.path.isdir(output_dir):
                os.makedirs(name=output_dir, exist_ok=True)
            if not os.path.isdir(output_dir):
                return False, False, False, ["Output directory did not exist or could not be created"]
        extension, raw_extension = determine_file_extension(input_path)
        vip_file = False
        if not extension:
            # Skip well known files with unsupported file extensions
            base_name = os.path.basename(input_path)
            if base_name in ["sitemap.xml", "favicon.ico", "robots.txt"]:
                vip_file = True
            else:
                return False, False, False, [f"Unknown extension '{raw_extension}' for input file"]
        handler = None
        is_binary = True
        if not vip_file:
            handler = self.handlers.get_handler_by_extension(extension)
            if not handler:
                return False, False, False, [f"Could not find handler for input file with extension {extension}"]
            is_binary = handler.is_binary()
        original_content = None
        try:
            with open(input_path, "rb") if is_binary else open(input_path, "r", encoding="utf-8") as file:
                original_content = file.read()
        except Exception as err:
            return False, False, False, [f"Could not load data from input file: {err}"]
        if not original_content:
            logger.warning(f"Input file '{input_path}' was empty")
        processed_content = None
        handler_errors = None
        was_copied = False
        do_copy = handler is None or vip_file or (handler.name() in self.handler_names and self.settings.get(f"disable_type_{handler.name()}", False)) or (extension in self.valid_extensions and self.settings.get(f"disable_suffix_{extension}", False))
        if do_copy:
            # Perform copy
            # logger.info(f"SUPPOSED TO COPY: {input_path} ({extension})")
            processed_content = original_content
            was_copied = True
        else:
            # logger.info(f"SUPPOSED TO PROCESS: {input_path} ({extension})")
            _processed_content, handler_errors = handler.process(original_content, input_path)
            processed_content = _processed_content
        # logger.info(f"Content of file {input_path} was of type {type(original_content)} while processed {type(processed_content)}")
        if handler_errors:
            return False, False, False, handler_errors
        if None == processed_content:
            logger.warning(f"Processed file '{input_path}' was empty")
        try:
            if not self.dry_run:
                with open(output_path, "wb") if is_binary else open(output_path, "w", encoding="utf-8") as file:
                    written = file.write(processed_content)
                    if written != len(processed_content):
                        return False, False, False, [f"Partially written output ({written} of {len(processed_content)} bytes)"]
        except Exception as err:
            return False, False, False, [f"Could not write data to output file: {err}"]
        try:
            if not self.dry_run:
                os.utime(output_path, (input_mtime, input_mtime))
        except Exception as err:
            return False, False, False, [f"Could not modify date of output file: {err}"]
        if self.gzip:
            try:
                gzip_path = f"{output_path}{gzip_suffix}"
                if not self.dry_run:
                    with open(gzip_path, "wb") as gzip_file:
                        gzip_content = zlib.compress(processed_content, level=9)
                        gzip_written = gzip_file.write(gzip_content)
                        if gzip_written != len(gzip_content):
                            return False, False, False, [f"Partially written gzip output ({gzip_written} of {len(gzip_content)} bytes)"]
            except Exception as err:
                return False, False, False, [f"Could not write gzip data to output file: {err}"]
            try:
                if not self.dry_run:
                    os.utime(gzip_content, (input_mtime, input_mtime))
            except Exception as err:
                return False, False, False, [f"Could not modify date of gzip file: {err}"]
        # All went well, go home happy!
        return True, was_copied, False, []

    def process_files_list_item_wrapper(self, item):
        input_path = item["input_path"]
        output_path = item["output_path"]
        single_start_time = datetime.datetime.now()
        result = {}
        # success, copied, skipped, message = self.process_file(input_path=input_path, output_path=output_path)
        result["success"], result["copied"], result["skipped"], result["messages"] = self.process_file(input_path=input_path, output_path=output_path)
        result["input_path"] = input_path
        result["output_path"] = output_path
        result["single_start_time"] = single_start_time
        single_end_time = datetime.datetime.now()
        single_interval = single_end_time - single_start_time
        result["single_end_time"] = single_end_time
        result["single_interval"] = single_interval
        result["was_slow"] = single_interval > datetime.timedelta(seconds=1)
        if result.get("was_slow"):
            result["messages"].extend([f"Processing of {result.get('input_path')} was slow ({human_delta(result.get('single_interval'))})"])
        return result

    def process_files_list(self, list):
        start_time = datetime.datetime.now()
        success_count = 0
        copied_count = 0
        skipped_count = 0
        messages = {}
        results = []
        if self.dry_run:
            logger.info(f"Dry run enabled, changes will NOT be commited to filesystem")
        if self.nproc > 1:
            logger.info(f"Doing multiprocessing with {self.nproc} cores")
            with Pool(self.nproc) as pool:
                results = pool.map(self.process_files_list_item_wrapper, list)
        else:
            logger.info(f"Doing sequential processing")
            for item in list:
                result = self.process_files_list_item_wrapper(item)
                results.append(result)
        # Summarize results
        for result in results:
            if result.get("success"):
                success_count += 1
            if result.get("copied"):
                copied_count += 1
            if result.get("skipped"):
                skipped_count += 1
            input_path = f"{result.get('input_path')}"
            for message in result.get("messages"):
                all = messages.get(message, [])
                messages[message] = [*all, input_path]
        end_time = datetime.datetime.now()
        interval = end_time - start_time
        failed = len(list) - success_count
        logger.info(f"Processing of {len(list)} files with {success_count} successful, {copied_count} copied and {skipped_count} skipped, generated {len(messages)} message(s) and took {human_delta(interval)} total")
        if messages:
            logger.warning(f"Messages encountered were:")
            for message, all in messages.items():
                logger.warning(f"{len(all)} x {message}")
                show_count = 5
                count = len(all)
                for index in range(min(show_count, count)):
                    logger.warning(f"    {all[index]}")
                if show_count < count:
                    logger.warning(f"    ... and {count-show_count} more")
                logger.warning("")
        return True

    def _process_existing_dir_to_existing_dir(self):
        input_paths = generate_file_list(self.input, tuple(self.valid_extensions))
        # logger.info(pprint.pformat(input_paths))
        list = []
        for input_path in input_paths:
            common = os.path.commonpath((os.path.abspath(self.output), os.path.abspath(input_path)))
            rel = os.path.relpath(os.path.abspath(input_path), os.path.abspath(self.input))
            output_path = os.path.join(os.path.abspath(self.output), rel)
            list.append({"input_path": input_path, "output_path": output_path})
        # logger.info(pprint.pformat(list))
        return list

    def _process_existing_dir_to_new_dir(self):
        # This is same as _process_existing_dir_to_existing_dir but with a mkdir first
        os.mkdir(self.output)
        self.output_exists = self.output and os.path.exists(self.output)
        return self._process_existing_dir_to_existing_dir()

    def _process_existing_dir_overwrite(self):
        input_paths = generate_file_list(self.input, tuple(self.valid_extensions))
        # logger.info(pprint.pformat(input_paths))
        list = []
        for input_path in input_paths:
            list.append({"input_path": input_path, "output_path": input_path})
        # logger.info(pprint.pformat(list))
        return list

    def process_files(self):
        not_implemented = "not implemented"
        if self.input_is_dir:
            if self.output:
                if self.output_is_dir:
                    if self.verbose:
                        logger.info(f"existing-dir-to-existing-dir easy peasy case")
                    list = self._process_existing_dir_to_existing_dir()
                    return self.process_files_list(list), None
                elif not self.output_exists:
                    if self.verbose or True:
                        logger.info(f"existing-dir-to-new-dir easy peasy case")
                    list = self._process_existing_dir_to_new_dir()
                    return self.process_files_list(list), None
                else:
                    return None, f"existing-dir-to-existing-file error"
            elif self.overwrite:
                if self.verbose or True:
                    logger.info(f"existing-dir-overwrite easy peasy case")
                list = self._process_existing_dir_overwrite()
                return self.process_files_list(list), None
            elif self.format:
                return None, f"format: {not_implemented}"
            else:
                return None, f"input dir specified without valid output option"
        else:
            if self.output:
                if self.output_is_dir:
                    return None, f"existing-file-to-existing-dir easy peasy case: {not_implemented}"
                elif not self.output_exists:
                    return None, f"existing-file-to-new-file easy peasy case: {not_implemented}"
                else:
                    if self.overwrite:
                        return None, f"existing-file-to-existing-file overwrite easy peasy case: {not_implemented}"
                    else:
                        return None, f"existing-file-to-existing-file non-overwrite error"
            elif self.overwrite:
                return None, "overwrite: {not_implemented} 2"
            elif self.format:
                return None, "format: {not_implemented} 2"
            else:
                return None, f"input file specified without valid output option 2"
